from dagsim.Node import _Node


class Selection(_Node):
    def __init__(self, name: str, function, args=None, kwargs=None, visible=True):
        super().__init__(name=name, function=function, args=args, kwargs=kwargs, visible=visible)

    @staticmethod
    def _build_object(**kwargs):
        # check params
        selection = Selection(**kwargs)
        return selection

    def _filter_output(self, output_dict):
        for key, value in output_dict.items():
            output_dict[key] = [value[i] for i in range(len(value)) if self.output[i]]
        return output_dict



from dagsim.Node import _Node, Node


class Missing(_Node):
    def __init__(self, name: str, underlying_value: Node, index_node: Node, visible=True):
        super().__init__(name=name, function=self._node_simulate, visible=visible)
        self.underlying_value = underlying_value
        self.parents = [underlying_value, index_node]
        self.index_node = index_node
        self.handle_multi_cols = underlying_value.handle_multi_cols
        self.handle_multi_return = underlying_value.handle_multi_return

    @staticmethod
    def _build_object(**kwargs):
        # check params
        missing = Missing(**kwargs)
        return missing

    def _node_simulate(self, *args):
        index_output = self.index_node.output
        output = [x if not y else 'NaN' for x, y in zip(self.underlying_value.output, index_output)]
        self.output = output



import numpy as np
import igraph as ig
import yaml
import os
from dagsim.utils._misc import parse_string_args


def create_functions_file(yaml_name: str):
    if not yaml_name.endswith(".yml"):
        yaml_name += ".yml"
    with open(yaml_name, 'r') as stream:
        yaml_file = yaml.safe_load(stream)

    file_str = ""

    nodes_dict = parse_string_args(yaml_file["graph"]["nodes"])
    print(nodes_dict)

    functions_dict = [val for val in nodes_dict.values() if "." not in val["function"]]
    print(functions_dict)

    for func in functions_dict:
        arguments = ", ".join(func["args"]) + ", ".join([f'{key}={val}' for key, val in func["kwargs"]])
        print(arguments)
        file_str += "def " + func + arguments + ":\n    pass\n\n\n"
    file_str = file_str[:-2]

    # add extension if it does not exist
    file_name = os.path.splitext(yaml_file["graph"]["python_file"])[0]+'.py'
    print(file_name)
    with open(file_name, 'w') as file:
        file.write(file_str)


def from_matrix(weight_matrix: np.ndarray, sem_type: str = "gauss", script_name: str = "script"):
    def create_function(name, weight_vector, sem_type):
        non_zero_indices = [i for i, e in enumerate(weight_vector) if e != 0]
        parents_names = ", ".join(["x" + str(i) for i in non_zero_indices])
        linear_equation = " + ".join(
            [str(weight_vector[e]) + " * x" + str(non_zero_indices[i]) for i, e in enumerate(non_zero_indices)])
        func_str = "def func_" + name + "(" + parents_names + "):\n\t" + name + " = "
        if sem_type not in ["logistic", "poisson"]:
            func_str += linear_equation + " + "
            if sem_type == "gauss":
                func_str += "np.random.normal(loc=0, scale=1)"
            elif sem_type == 'exp':
                func_str += "np.random.exponential(loc=0, scale=1)"
            elif sem_type == 'gumbel':
                func_str += "np.random.gumbel(loc=0, scale=1)"
            elif sem_type == 'uniform':
                func_str += "np.random.uniform(low=-1, high=1)"
        elif sem_type == 'logistic':
            func_str += "np.random.binomial(1, sigmoid(" + linear_equation + ")) * 1.0"
        elif sem_type == 'poisson':
            func_str += "np.random.poisson(1, np.exp(" + linear_equation + ")) * 1.0"
        else:
            raise ValueError('unknown sem type')
        func_str += "\n\treturn " + name

        return func_str

    def create_node(name: str, parents: list = None):
        # args_str = {"x" + str(i): "Node_x" + str(i) for i in parents}
        node_def = "Node_" + name + " = ds.Node(name='" + name + "', function="
        if parents:
            arg_string = "{"
            for i in parents:
                arg_string += "'x" + str(i) + "': Node_" + "x" + str(i) + ", "
            arg_string = arg_string[:-2] + "}"
            node_def += "func_" + name + ", kwargs=" + arg_string + ")"
        else:
            node_def += "np.random.normal, kwargs={'loc': 0, 'scale': 1})"
        return node_def

    assert weight_matrix.shape[0] == weight_matrix.shape[1], "The weight matrix should be a square matrix"

    G = ig.Graph.Weighted_Adjacency(weight_matrix.tolist())
    top_order = G.topological_sorting()

    imports = "import dagsim.base as ds \n" \
              "import numpy as np \n"

    if sem_type == "logistic":
        imports += "from scipy.special import expit as sigmoid\n"

    imports += "\n\n"

    functions = ""
    for index, column in enumerate(weight_matrix.T):
        if any(column):
            functions += create_function("x" + str(index), column, sem_type)
            functions += "\n\n\n"

    nodes = ""
    for node_index in top_order:
        if any(weight_matrix.T[node_index]):
            non_zero_indices = [i for i, e in enumerate(weight_matrix.T[node_index]) if e != 0]
            nodes += create_node("x" + str(node_index), non_zero_indices)
            nodes += "\n"
        else:
            nodes += create_node("x" + str(node_index))
            nodes += "\n"

    graph_def = "\nlistNodes = ["
    for node_index in top_order:
        graph_def += "Node_x" + str(node_index) + ", "
    graph_def = graph_def[:-2] + "]\n"
    graph_def += "graph = ds.Graph('myGraph', listNodes) \n"

    script = imports + functions + nodes + graph_def
    with open(script_name + ".py", "w") as file:
        file.write(script)


def from_csv(file_name: str, sem_type: str, script_name: str):
    weight_matrix = np.genfromtxt(file_name+".csv")
    from_matrix(weight_matrix, sem_type=sem_type, script_name=script_name)


if __name__ == "__main__":
    # matrix_example = np.array([[0, 0, 2, 1], [0, 0, 3, 0], [0, 0, 0, 1], [0, 0, 0, 0]])
    #
    # from_matrix(matrix_example, sem_type="logistic", script_name="gaussDagSim")

    create_functions_file("testyml.yml")



def find_order(list_of_plates: list):
    """ A function that finds which plates are sub-plates of other plates.

    Args:
        list_of_plates (list): A list of lists, where each sublist contains the nodes found in the plate with
            the corresponding index.

    Returns:
        dict: A dictionary of the sub-plates, if any, of each plate.

    """
    resDict = {k: [] for k in range(1, len(list_of_plates) + 1)}
    for plate in list_of_plates:
        if len(plate) != 1:
            checkList = [inPlate for inPlate in list_of_plates if len(inPlate) < len(plate)]
            for inPlate in checkList:
                if set(inPlate).issubset(set(plate)):
                    resDict[list_of_plates.index(plate) + 1].append(list_of_plates.index(inPlate) + 1)
    return resDict


def process_dict(redun_dict: dict):
    """ A function that finds the direct child plates, if any, of each plate.

    Args:
        redun_dict (dict): A dictionary containing the, possibly redundant, sub-plates of each plate.

    Returns:
        dict: A dictionary of the direct child plates, if any, of each plate.

    """
    innder_plateIdxs = set([item for sublist in redun_dict.values() for item in sublist])
    for plate in innder_plateIdxs:
        keysPlate = [key for key in redun_dict.keys() if plate in redun_dict[key]]
        if len(keysPlate) == 1:
            continue
        for childPlate in keysPlate:
            for parentPlate in keysPlate:
                if childPlate in redun_dict[parentPlate]:
                    redun_dict[parentPlate].remove(plate)
            if len(keysPlate) == 1:
                break
    return redun_dict


def transform(processed_dict, root_plates, subs_dict):
    """ A function that generates the .dot string for all the plates in the graph, in the correct order.

    Args:
        processed_dict (dict): A dictionary of the direct child plates, if any, of each plate.
        root_plates (list): A list of the root plates.
        subs_dict (dict): A dictionary of strings to define each plate in .dot format.

    Returns:
        string: A string defining all the plates and their nodes, in .dot format.

    """
    dot_string = ""
    level = 0

    def plate_string(processed_dict, root_plates):
        nonlocal dot_string, level, subs_dict
        for idx, root in enumerate(root_plates):
            dot_string = dot_string + subs_dict[root] + "\n"
            if not processed_dict[root]:
                if idx == (len(root_plates) - 1):
                    level -= 1
                    dot_string = dot_string + "}"
            else:
                level += 1
                plate_string(processed_dict, processed_dict[root])
                dot_string = dot_string[:-1]
            dot_string = dot_string + "}"

    plate_string(processed_dict, root_plates)
    return dot_string[0:-1]


def find_roots(processed_dict):
    """ A function that finds the plates that are not contained in other plates.

    Args:
        processed_dict (dict): A dictionary of the direct child plates, if any, of each plate.

    Returns:
        list: A list of the root plates.

    """
    root_plates = list(range(1, len(processed_dict.keys()) + 1))
    for key in processed_dict.keys():
        for val in processed_dict[key]:
            root_plates.pop(root_plates.index(val))
    return root_plates


def get_plate_dot(plate_embedding):
    """ A wrapper function that generates the part of the .dot string corresponding to the plates.

    Args:
        plate_embedding (dict): A dictionary containing the label and nodes of each plate.

    Returns:
        str: A string defining the plates in .dot format.

    """
    subs_dict = {
        i: "subgraph cluster_" + str(i) + " {" + ", ".join(plate_embedding[i][1]) + ';\n label="' +
           plate_embedding[i][0] + '"' + ';\n labelloc=b' for i in range(1, len(plate_embedding.keys()))}
    ordered_dict = find_order([plate_embedding[i][1] for i in range(1, len(plate_embedding.keys()))])
    processed_dict = process_dict(ordered_dict)
    root_plates = find_roots(processed_dict)
    dot_string = transform(processed_dict, root_plates, subs_dict)
    return dot_string


if __name__ == "__main__":
    sample_plate_embedding = {0: (None, ['Prior', 'Node4']), 1: ('plate_1', ['Node1', 'Node2']),
                              2: ('plate_2', ['Node5', 'Node1', 'Node2'])}
    # Note that the 0th plate is excluded from the .dot string generation.
    print(get_plate_dot(sample_plate_embedding))



import unittest

from dagsim.base import Graph, Node, Missing
import random


class TestMissing(unittest.TestCase):

    def setUp(self) -> Graph:

        def get_index(counter=[0]):
            index_list = [False, True, False, True, False]
            counter[0] += 1
            return index_list[counter[0] - 1]

        def get_val(counter=[0]):
            index_list = [1, 2, 3, 4, 5]
            counter[0] += 1
            return index_list[counter[0] - 1]

        index = Node("i", function=get_index)
        toBeMissed = Node(name="toBeMissed", function=get_val)
        missingNode = Missing(name="M1", underlying_value=toBeMissed, index_node=index)
        my_graph = Graph(name="graph1", list_nodes=[index, toBeMissed, missingNode])
        return my_graph

    def test_missing(self):
        data = self.setUp().simulate(5)
        self.assertEqual([False, True, False, True, False], data["i"])
        self.assertEqual([1, 2, 3, 4, 5], data["toBeMissed"])
        self.assertEqual([1, 'NaN', 3, 'NaN', 5], data['M1'])


if __name__ == '__main__':
    unittest.main()



def parse_string_args(nodes):
    # For each node, separate the function's name from its arguments, if not separated already
    for key in nodes.keys():
        if "type" in nodes[key] and nodes[key]["type"] == "Missing":
            continue
        else:
            nodes[key] = check_node_syntax(nodes[key])
            if "type" not in nodes[key]:
                nodes[key]["type"] = "Node"
            nodes[key]["args"] = []
            if "kwargs" not in nodes[key]:
                nodes[key]["kwargs"] = {}

            if "(" in nodes[key]["function"]:
                if nodes[key]["kwargs"]:
                    raise SyntaxError(
                        "Using a python-like definition, i.e. with parantheses, with separate kwargs is not "
                        "allowed. Use one way or the other.")
                elif "()" in nodes[key]["function"]:
                    nodes[key]["function"] = nodes[key]["function"][:-2]
                else:
                    nodes[key]["function"], nodes[key]["args"], nodes[key]["kwargs"] = split_func_and_args(
                        nodes[key]["function"])
    return nodes


def split_func_and_args(func_expression: str):
    # Split a string of the form "func_name(arg1, arg2,.., kwarg1=val1, kwarg2=val2,..)" into
    # func_name, [arg1, arg2,..], {kwarg1=val1, kwarg2=val2,..}
    func_expression = func_expression.replace(" ", "")
    inputs = func_expression[func_expression.find("(") + 1: func_expression.find(")")]
    first_kwarg_index = check_args_order(inputs)
    inputs = inputs.split(",")
    args = inputs[:first_kwarg_index]
    for arg_idx in range(len(args)):
        args[arg_idx] = get_number(args[arg_idx]) if is_numeric(args[arg_idx]) else args[arg_idx]
    inputs = inputs[first_kwarg_index:]
    kwargs = {}
    for kwarg in inputs:
        arg_name = kwarg[:kwarg.find("=")]
        kwargs[arg_name] = kwarg[kwarg.find("=") + 1:]
        kwargs[arg_name] = get_number(kwargs[arg_name]) if is_numeric(kwargs[arg_name]) else kwargs[arg_name]
    func_name = func_expression[:func_expression.find("(")]
    return func_name, args, kwargs


def get_number(string: str):
    num = float(string)
    if num == int(num):
        num = int(num)
    return num


def is_numeric(string: str):
    try:
        float(string)
        return True
    except ValueError:
        return False


def check_args_order(all_args_str: str):
    # Check that no positional args come after kwargs
    all_args_str = all_args_str.split(",")
    first_kwarg_index = next((all_args_str.index(x) for x in all_args_str if "=" in x), None)
    if first_kwarg_index is not None:
        for kwargs in range(first_kwarg_index, len(all_args_str)):
            if "=" not in all_args_str[kwargs]:
                raise RuntimeError("Positional argument after keyword argument")
    else:
        first_kwarg_index = len(all_args_str)
    return first_kwarg_index


def check_node_syntax(node):
    if not isinstance(node, dict):
        assert isinstance(node, str), "One line definitions should contain the function definition itself."
        node_dict = {"function": node}
        return node_dict
    else:
        return node



import unittest
from typing import List

from dagsim.Node import Node
from dagsim.base import Node, Graph


class TestPlates(unittest.TestCase):

    def add(self, a, b):
        return a + b

    def setUp(self) -> list[Node]:
        n0 = Node("n0", function=int, args=[1])
        n1 = Node("n1", function=int, args=[1], plates=["1"])
        n2 = Node("n2", function=self.add, args=[n0, n1], plates=["2"])
        n3 = Node("n3", function=lambda x: x, args=[n1])
        return [n0, n1, n2, n3]

    def test_child_in_plate_no_rep(self):
        graph = Graph(self.setUp()[:2])
        data = graph.simulate(2)
        self.assertEqual(data["n0"], [1, 1])
        self.assertEqual(data["n1"], [1, 1])

    def test_child_in_plate_with_rep(self):
        graph = Graph(self.setUp()[:2], plates_reps={"1": 2})
        data = graph.simulate(2)
        self.assertIn("n1_0_", data)
        self.assertIn("n1_1_", data)
        self.assertEqual(data["n0"], [1, 1])
        self.assertEqual(data["n1_1_"], [1, 1])

    def test_parent_in_plate(self):
        nodes = self.setUp()
        n1 = nodes[1]
        n3 = nodes[3]
        graph = Graph([n1, n3], plates_reps={"1": 2})
        data = graph.simulate(2)
        self.assertIn("n1_1_", data)
        self.assertEqual(data["n1_1_"], [1, 1])
        self.assertEqual(data["n3"], [[1, 1], [1, 1]])


if __name__ == '__main__':
    unittest.main()



import yaml
from dagsim.base import Graph, Node, Selection, Stratify, Missing
from dagsim.utils._misc import parse_string_args
from inspect import getmembers, isfunction
import importlib
import pandas as pd
import igraph as ig
import os.path


class DagSimSpec:
    def __init__(self, file_name: str):
        self.functions_list = None
        self.top_order = []
        self.graph = None
        with open(file_name, 'r') as stream:
            self.yaml_file = yaml.safe_load(stream)
        # todo check parsing floats or integers

    def parse(self, verbose: bool = True, draw: bool = True):

        nodes_dict = parse_string_args(self.yaml_file["graph"]["nodes"])

        adj_matrix, node_names = self._build_adj_matrix(nodes_dict)

        assert self._check_acyclicity(adj_matrix), "The graph is not acyclic."

        self._find_top_order(adj_matrix, node_names)

        try:
            python_file = self.yaml_file["graph"]["python_file"]
            assert python_file.endswith(".py"), "Please use a proper python file."
            assert os.path.isfile(python_file), "The file \"" + python_file + "\" doesn't exist."
            python_file = python_file[:-3]

            functions_file = importlib.import_module(python_file)
            self.functions_list = getmembers(functions_file, isfunction)
        except KeyError:
            self.functions_list = []

        self._build_graph_from_nodes(nodes_dict)
        if verbose:
            print(self.graph)
        if draw:
            self.graph.draw()
        data = self._simulate_data()
        return data

    def _build_adj_matrix(self, nodes: dict):
        node_names = nodes.keys()
        # allow for non-boolean Missing-index output
        parents_dict = {k: [v for v in (list(nodes[k]["kwargs"].values()) + nodes[k]["args"]) if
                            v in node_names] for k in node_names if nodes[k]["type"] != "Missing"}
        parents_dict.update({k: [nodes[k]["underlying_value"], nodes[k]["index_node"]] for k in node_names if
                             nodes[k]["type"] == "Missing"})

        pd_dict = pd.DataFrame(0, columns=node_names, index=node_names)
        for child in node_names:
            for parent in parents_dict[child]:
                pd_dict[child][parent] = 1
        return pd_dict.to_numpy(), list(node_names)

    def _check_acyclicity(self, adj_mat):
        pdDAGGraph = ig.Graph.Weighted_Adjacency(adj_mat.tolist())
        return ig.Graph.is_dag(pdDAGGraph)

    def _find_top_order(self, adj_mat, names):
        G = ig.Graph.Weighted_Adjacency(adj_mat.tolist())
        top_order = G.topological_sorting()
        top_order = [names[i] for i in top_order]
        self.top_order = top_order

    def _build_graph_from_nodes(self, nodes: dict):
        self.list_nodes = []
        for key in self.top_order:
            node_type = self._get_node_type(nodes[key])
            nodes[key] = {**nodes[key], **{"name": key}}
            if node_type == "Missing":
                node = self._build_missing_node(nodes[key])
            else:
                node = self._build_other_node(nodes[key], node_type)
            self.list_nodes.append(node)

        if "name" not in self.yaml_file["graph"]:
            self.yaml_file["graph"]["name"] = "Graph"

        plate_reps = self._get_plates_reps()

        self.graph = Graph(name=self.yaml_file["graph"]["name"], list_nodes=self.list_nodes, plates_reps=plate_reps)

    def _build_missing_node(self, node: dict) -> Missing:
        assert "underlying_value" in node, "'underlying_value' is not specified"
        assert "index_node" in node, "'index_node' is not specified"

        node["underlying_value"] = self._get_node_by_name(node["underlying_value"])
        node["index_node"] = self._get_node_by_name(node["index_node"])

        return Missing(**node)

    def _build_other_node(self, node, node_type):
        node["function"] = self._get_func_by_name(self.functions_list, node["function"])
        node["kwargs"] = {
            k: self._get_node_by_name(v) if self._get_node_by_name(v) is not None else v for
            k, v in node["kwargs"].items()}

        node["args"] = [
            self._get_node_by_name(v) if self._get_node_by_name(v) is not None else v for v
            in node["args"]]

        node = self._clear_strs(node)

        if node_type == "Node":
            if "plates" in node:
                node["plates"] = str(node.get("plates")).replace(" ", "").split(",")
            node = Node._build_object(**node)

        elif node_type == "Selection":
            node = Selection._build_object(**node)

        elif node_type == "Stratify":
            node = Stratify._build_object(**node)

        return node

    def _get_node_type(self, node):
        node_type = node.get("type")  # always not None because parse_string_args adds the type
        node.pop("type")
        if node_type in ["Node", "Selection", "Missing", "Stratify"]:
            return node_type
        else:
            raise TypeError("\"" + node_type + "\" is not a valid node type. \"type\" should be either Node, "
                                               "Selection, or Stratify.")

    def _get_func_by_name(self, functions_list: list, func_name: str):
        for name, func in functions_list:
            if name == func_name:
                return func
        try:
            func = self._get_implicit_func(func_name=func_name)
            return func
        except (AttributeError, ModuleNotFoundError):
            raise ImportError("Couldn't find the function \"" + func_name + "\"")

    def _get_implicit_func(self, func_name: str):
        # For functions not defined in the python file, but rather coming from external libraries, such as numpy.
        first_part = func_name.rfind(".")
        module_name = func_name[:first_part]
        func_name = func_name[first_part + 1:]
        module = importlib.import_module(module_name)
        func = getattr(module, func_name)
        return func

    def _clear_strs(self, node):
        for ind in range(len(node["args"])):
            if isinstance(node["args"][ind], str):
                if node["args"][ind].startswith(("'", '"')):
                    node["args"][ind] = node["args"][ind][1:-1]
        for k, v in node["kwargs"].items():
            if isinstance(v, str):
                if v.startswith(("'", '"')):
                    node["kwargs"][k] = v[1:-1]
        return node

    def _get_plates_reps(self):
        plate_dict = self.yaml_file["graph"].get("plates_reps")
        if plate_dict is not None:
            plate_dict = {str(key): val for key, val in plate_dict.items()}
        return plate_dict

    def _get_node_by_name(self, name):
        try:
            node = [i for i in self.list_nodes if i.name == name][0]
        except (ValueError, IndexError):
            node = None
        return node

    def _simulate_data(self):
        data = self.graph.simulate(**self.yaml_file["instructions"]["simulation"])
        return data


if __name__ == "__main__":
    # parser = DagSimSpec(file_name="test_yaml_missing.yml")
    parser = DagSimSpec(file_name="testyml2.yml")
    data = parser.parse()
    print(data)



import dagsim.base as ds 
import numpy as np 
from scipy.special import expit as sigmoid


def func_x2(x0, x1):
	x2 = np.random.binomial(1, sigmoid(2 * x0 + 3 * x1)) * 1.0
	return x2


def func_x3(x0, x2):
	x3 = np.random.binomial(1, sigmoid(1 * x0 + 1 * x2)) * 1.0
	return x3


Node_x0 = ds.Node(name='x0', function=np.random.normal, kwargs={'loc': 0, 'scale': 1})
Node_x1 = ds.Node(name='x1', function=np.random.normal, kwargs={'loc': 0, 'scale': 1})
Node_x2 = ds.Node(name='x2', function=func_x2, kwargs={'x0': Node_x0, 'x1': Node_x1})
Node_x3 = ds.Node(name='x3', function=func_x3, kwargs={'x0': Node_x0, 'x2': Node_x2})

listNodes = [Node_x0, Node_x1, Node_x2, Node_x3]
graph = ds.Graph('myGraph', listNodes) 
data = graph.simulate(4)


import numpy as np


def get_args_str(doc_string, n):
    start = doc_string.find("(")
    end = doc_string.find(",")
    while end >= 0 and n > 1:
        end = doc_string.find(",", end + 1)
        n -= 1
    if end == -1:
        end = doc_string.find(")")
    return doc_string[start + 1:end]


def get_args_list(args_str):
    pass


def check_for_numeric_type(string: str):
    num = float(string)
    if num == int(num):
        num = int(num)
    return num


def is_numeric(string: str):
    try:
        float(string)
        return True
    except ValueError:
        return False


def get_args_st(doc, n):
    start = doc.find("(")
    end = doc.find(")")
    doc = doc[doc.find("("): doc.find(")")]
    feq = doc.find("=")
    print(doc[:feq])


if __name__ == "__main__":
    # doc = "fds(dasd,dasd dasd) dasd sdaftqwe "
    # doc = np.add.__doc__
    # args = get_args_st(doc, 2)
    # print(args)
    is_numeric("2")
    is_numeric("2a")
    is_numeric("-2")
    is_numeric("-2.2")
    is_numeric("23.")



import unittest

from dagsim.base import Graph, Node
import numpy as np


class TestVectorize(unittest.TestCase):

    def setUp(self) -> Graph:
        def get_normal(counter=[0]):
            index_list = [0, 1, 2, 3, 4]
            counter[0] += 1
            return index_list[counter[0] - 1]

        Normal = Node(name="Normal", function=get_normal)
        Vectorized = Node(name="Vectorized", function=np.arange, kwargs={"start": 0}, size_field="stop")
        my_graph = Graph(name="graph1", list_nodes=[Normal, Vectorized])
        return my_graph

    def test_vectorized(self):
        data = self.setUp().simulate(5)
        print(data)
        self.assertEqual([0, 1, 2, 3, 4], data["Normal"])
        self.assertEqual([0, 1, 2, 3, 4], list(data["Vectorized"]))
        self.assertEqual(data["Normal"], list(data["Vectorized"]))


if __name__ == '__main__':
    unittest.main()



from inspect import getfullargspec


class _Node:
    def __init__(self, name: str, function, plates=None, observed=True, args: list = None, kwargs: dict = None,
                 size_field=None, visible=True):
        if kwargs is None:
            kwargs = {}
        if args is None:
            args = []
        if plates is None:
            plates = []
        self.name = name
        self._constructor = {"args": args, "kwargs": kwargs}
        self._args, self._kwargs = self._parse_func_arguments()
        self.parents = []
        self._update_parents()
        self.function = function
        self.output = None
        self.observed = observed
        self.visible = visible
        self.plates = plates
        self.size_field = size_field

    def _parse_func_arguments(self):
        args, kwargs = self._constructor.values()
        args = [
            (lambda x: (lambda index: x.output[index]))(a) if isinstance(a, Node) else (lambda x: (lambda index: x))(
                a) for a in args]
        kwargs = dict(
            [(k, (lambda x: (lambda index: x.output[index]))(v)) if isinstance(v, Node) else (
                k, (lambda x: (lambda index: x))(v)) for k, v in kwargs.items()])
        return args, kwargs

    def _get_func_args(self, index):
        return [a(index) for a in self._args]

    def _get_func_kwrgs(self, index, output_path):
        d = dict([(k, v(index)) for k, v in self._kwargs.items()])
        try:  # This would throw a TypeError in case the function is ufunc
            if "output_path" in getfullargspec(self.function).args:
                d["output_path"] = output_path
        except TypeError:
            pass
        return d

    def __str__(self):
        main_str = ""
        main_str += "\tname: " + self.name + "\n"
        main_str += "\ttype: " + self.__class__.__name__ + "\n"
        main_str += "\tfunction: " + self.function.__name__ + "\n"
        if self.parents:
            main_str += "\tparents: " + ", ".join([par.name for par in self.parents]) + "\n"
        else:
            main_str += "\tparents: None"
        return main_str

    def _forward(self, idx, output_path):
        return self.function(*self._get_func_args(idx), **self._get_func_kwrgs(idx, output_path))

    def _node_simulate(self, num_samples, output_path):
        if self.size_field is None:
            self.output = [self._forward(i, output_path) for i in range(num_samples)]
        else:
            self.output = self._vectorize_forward(num_samples, output_path)

    def _vectorize_forward(self, size, output_path):
        #  Args and KwArgs would be used once. Index just for convenience
        return self.function(*self._get_func_args(0), **self._get_func_kwrgs(0, output_path),
                             **{self.size_field: size})  # .tolist()

    def __len__(self):
        return len(self.parents)

    def _update_parents(self):
        self.parents = list(set([v for v in self._constructor["args"] if isinstance(v, Node)] + list(
            v for v in self._constructor["kwargs"].values() if isinstance(v, Node))))


class Node(_Node):
    def __init__(self, name: str, function, args=None, kwargs=None, plates=None, size_field=None, observed=True,
                 visible=True,
                 handle_multi_cols=False, handle_multi_return=None):
        super().__init__(name=name, function=function, args=args, kwargs=kwargs,
                         plates=plates, observed=observed, visible=visible, size_field=size_field)
        self.handle_multi_cols = handle_multi_cols
        self.handle_multi_return = handle_multi_return

    @staticmethod
    def _build_object(**kwargs):
        # check params
        generic = Node(**kwargs)
        return generic



import dagsim.base as ds
import numpy as np


def square(arg):
    return arg * arg


def main():

    X = ds.Node(name="X", function=np.random.normal)
    Y = ds.Node(name="Y", function=square, kwargs={"arg": X})
    graph = ds.Graph(name="demo_graph", list_nodes=[X, Y])

    data = graph.simulate(num_samples=2)
    print("Simulated data", data)


if __name__ == "__main__":
    main()



import unittest

from dagsim.base import Graph, Node, Stratify
import random


class TestStratify(unittest.TestCase):

    def setUp(self) -> Graph:
        def get_normal(counter=[0]):
            index_list = [0, 1, 2, 3, 4]  # should have 5 even numbers
            counter[0] += 1
            return index_list[counter[0] - 1]

        def stratify(val):
            if val % 2 == 0:
                return "Even"
            else:
                return "Odd"

        Normal = Node(name="Normal", function=get_normal)
        StratifyNode = Stratify(name="Stratify", function=stratify, kwargs={"val": Normal})
        my_graph = Graph(name="graph1", list_nodes=[Normal, StratifyNode])
        return my_graph

    def test_stratify(self):
        data = self.setUp().simulate(5, stratify=True)
        self.assertEqual({'Normal': [1, 3]}, data["Odd"])
        self.assertEqual({'Normal': [0, 2, 4]}, data["Even"])


if __name__ == '__main__':
    unittest.main()



import numpy as np
# from numpy.random import normal
import random


def add(params0, params1):
    return params0 + params1


def add_noise(x):
    return x + np.random.normal()


def printing(num, phra):
    aa = phra * num
    return aa


def get_bool():
    return random.choice([False, True])


def square_plus_constant(z, constant):
    # print(globals())
    # print(globals()["testValue"])
    # print(testValue)
    return np.square(z) + constant


def double(param, add1, add2):
    return np.square(param) + add1 - add2


# def normal(**kwargs):
#     return np.random.normal(**kwargs)



import unittest

from dagsim.base import Graph, Node, Selection
import random


class TestSelection(unittest.TestCase):

    def setUp(self) -> Graph:
        def get_normal(counter=[0]):
            index_list = [0, 1, 2, 3, 4, 5, 6, 7, 8]  # should have 5 even numbers
            counter[0] += 1
            return index_list[counter[0] - 1]

        def select(val):
            if val % 2 == 0:
                return True
            else:
                return False

        Normal = Node(name="Normal", function=get_normal)
        Select = Selection(name="Select", function=select, kwargs={"val": Normal})
        my_graph = Graph(name="graph1", list_nodes=[Normal, Select])
        return my_graph

    def test_selection(self):
        data = self.setUp().simulate(5)
        self.assertEqual([0, 2, 4, 6, 8], data["Normal"])


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages


setup(
    name='dagsim',
    version='1.0.7',
    packages=find_packages(),
    url='https://github.com/uio-bmi/dagsim',
    license='AGPL-3.0 License',
    author='Ghadi S. Al Hajj',
    author_email='ghadia@uio.no',
    description='A framework and specification language for simulating data based on user-defined graphical models',
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=['graphviz>=0.16',
                      'numpy>=1.20.2',
                      'pandas>=1.2.4',
                      'igraph>=0.9.6',
                      'pyyaml',
                      'ipython>=7.27.0'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'dagsim-quickstart = dagsim.utils.quickstart:main'
        ]
    },
)



from dagsim.Node import _Node


class Stratify(_Node):
    def __init__(self, name: str, function, args=None, kwargs=None, visible=True):
        super().__init__(name=name, function=function, args=args, kwargs=kwargs, visible=visible)

    @staticmethod
    def _build_object(**kwargs):
        # check params
        stratify = Stratify(**kwargs)
        return stratify

    def _filter_output(self, output_dict):
        node_names = output_dict.keys()
        strata = list(set(self.output))
        # A dictionary of dictionaries. Outer dictionary with keys=strata, and inner dictionaries with keys=node_names
        new_dict = {key: {node: [] for node in node_names} for key in strata}

        for i, stratum in enumerate(self.output):
            for k, v in output_dict.items():
                new_dict[stratum][k].append(v[i])

        return new_dict


import numpy as np
from typing import Union
from graphviz import Source
import pandas as pd
import igraph as ig
from dagsim.utils.processPlates import get_plate_dot
import time
import datetime
import copy
from dagsim.Node import Node
from dagsim.Missing import Missing
from dagsim.Selection import Selection
from dagsim.Stratify import Stratify


class Graph:
    def __init__(self, list_nodes, name="Graph", plates_reps: dict = None):
        self._check_args(list_nodes)
        self.name = name
        self.nodes = list_nodes  # [None] * num_nodes
        self.plates = {}
        self.adj_mat = pd.DataFrame()
        self.top_order = []
        self._update_topol_order()
        self._update_plate_embedding()
        self.folded_dot_str = self._generate_dot()
        if plates_reps is not None:
            self.unfolded_dot_str = ""
            self.unfold_graph(plates_reps)

    def unfold_graph(self, reps):
        removed_nodes = self._replicate_nodes(reps)  # When a node is replicated, the original one is removed
        # todo reserve _agg names if there are plates in the graph
        self._update_topol_order()
        self._update_nodes(removed_nodes)
        self._update_topol_order()
        self._update_plate_embedding()
        self.unfolded_dot_str = self._generate_dot()

    @staticmethod
    def _check_args(list_nodes):
        assert len([select for select in list_nodes if isinstance(select, Selection)]) <= 1, "A graph can have at " \
                                                                                             "most one Selection node. "
        assert len([strat for strat in list_nodes if isinstance(strat, Stratify)]) <= 1, "A graph can have at most " \
                                                                                         "one Stratify node. "

    def _get_nodes_to_aggregate(self):
        # Returns all the nodes that are in plates if their children are outside that plate.
        # This is used to check whether to aggregate the replications or not.
        nodes_to_aggregate = []
        for node in self.nodes:
            parents = node.parents
            for parent in parents:
                node_plates = set(node.plates)
                parent_plates = set(parent.plates)
                if not node_plates & parent_plates:
                    nodes_to_aggregate.append(parent)
        return nodes_to_aggregate

    def _replicate_nodes(self, plates_reps):
        # todo check if plates and not plate_reps
        # Replicates all the nodes found in plates based on plates_reps. Also set 'plates' to None on these nodes.
        parents_to_aggregate = self._get_nodes_to_aggregate()
        nodes_to_remove = []
        new_nodes = []
        for node in self.nodes:
            if node.plates:
                nodes_to_remove.append(node)
                node_replicas = [copy.deepcopy(node) for _ in range(plates_reps[node.plates[0]])]
                new_nodes.extend(node_replicas)
                for i in range(len(node_replicas)):
                    setattr(node_replicas[i], "name", node.name + f'_{i}_')
                if node in parents_to_aggregate:
                    new_nodes.append(
                        Node(name=f'{node.name}_agg', function=lambda *x: list(x), args=node_replicas, observed=False))
        self.nodes.extend(new_nodes)
        self.nodes = list(set(self.nodes) - set(nodes_to_remove))
        # deepcopy will copy the parent with the same name to different references, so checking for name solves it
        # copy updates all copies upon updating any of them, so keep the names instead:
        removed_nodes_names = [rmnode.name for rmnode in nodes_to_remove]
        return removed_nodes_names

    def __str__(self):
        main_str = ""
        for idx, node in enumerate(self.nodes):
            main_str += "Node " + str(idx + 1) + ":\n"
            main_str += node.__str__() + "\n"
        return main_str[:-1]

    def _update_plate_embedding(self):
        def get_key_by_label(label):
            for key in plateDict.keys():
                if plateDict[key][0] == label:
                    return key

        plateDict = {0: (None, [])}
        idx = 1
        labels = []
        for node in self.nodes:
            if node.plates:
                for label in node.plates:
                    if label not in labels:
                        plateDict[idx] = (label, [])
                        labels.append(label)
                        idx += 1
                    plateDict[get_key_by_label(label)][1].append(node.name)
            else:
                plateDict[0][1].append(node.name)
        self.plates = plateDict

    def _add_node(self, node: Union[Node, Selection, Stratify, Missing]):
        if node not in self.nodes:
            self.nodes.append(node)
        # update the topological order whenever a new node is added
        self._update_topol_order()

    def _get_selection(self):
        # todo change to missing way
        check_for_selection = next((item for item in self.nodes if item.__class__.__name__ == "Selection"), None)
        if check_for_selection is not None:
            return self.nodes.index(check_for_selection)
        else:
            return None

    def _get_stratify(self):
        check_for_stratify = next((item for item in self.nodes if item.__class__.__name__ == "Stratify"), None)
        if check_for_stratify is not None:
            return self.nodes.index(check_for_stratify)
        else:
            return None

    def _get_missing(self):
        check_for_missing = [item for item in self.nodes if item.__class__.__name__ == "Missing"]
        return check_for_missing if check_for_missing else None

    def get_node_by_name(self, name: str):
        if not isinstance(name, str):
            return None
        else:
            node = next((item for item in self.nodes if item.name == name), None)
            if node is None:
                print(name, " not found")
            return node

    def _update_adj_mat(self):
        nodes_names = [node.name for node in self.nodes]
        matrix = pd.DataFrame(data=np.zeros([len(self.nodes), len(self.nodes)]), dtype=int,
                              columns=nodes_names,
                              index=nodes_names)
        for node in self.nodes:
            if node.parents is not None:
                for parent in node.parents:
                    matrix[node.name][parent.name] = 1
        self.adj_mat = matrix

    def _update_topol_order(self):
        # todo reorder in chronological order after the topological one
        self._update_adj_mat()
        G = ig.Graph.Weighted_Adjacency(self.adj_mat.to_numpy().tolist())
        top_order = G.topological_sorting()
        top_order = [list(self.adj_mat.columns)[i] for i in top_order]
        self.top_order = top_order

    def __getitem__(self, item):
        return self.nodes[item]

    def __len__(self):
        return len(self.nodes)

    def _generate_dot(self):

        shape_dict = {'Node': "ellipse", 'Selection': "doublecircle", 'Stratify': "doubleoctagon",
                      "Missing": "Mcircle"}
        dot_str = 'digraph G{\n'
        # add the visible nodes
        for node_idx in range(len(self)):
            if self[node_idx].visible:
                my_str = '"%s"' % self[node_idx].name
                my_str += " [shape=" + shape_dict[type(self[node_idx]).__name__] + "];\n"
                dot_str = dot_str + my_str

        # add the edges of both vertices are visible
        for parent_node in self.adj_mat:
            if self.get_node_by_name(parent_node).visible:
                for node_idx in self.adj_mat.loc[parent_node].index:
                    if self.adj_mat.loc[parent_node].loc[node_idx] == 1:
                        if self.get_node_by_name(node_idx).visible:
                            tmp_str = '"%s" -> "%s";\n' % (parent_node, node_idx)
                            dot_str += tmp_str

        # check if there are any plates defined in the graph
        if len(self.plates) > 1:
            dot_str += get_plate_dot(self.plates)
        dot_str += "}"
        return dot_str

    def draw(self, folded=True):
        dot_str = self.folded_dot_str if folded else self.unfolded_dot_str
        with open(self.name + "_DOT.txt", "w") as file:
            file.write(dot_str)

        s = Source(dot_str, filename=self.name, format="png")
        try:
            s.view(cleanup=True, quiet_view=True)
        except (TypeError, FileNotFoundError):
            from IPython.display import display
            s.render()
            display(Source(dot_str))

    def simulate(self, num_samples, csv_name="", output_path="./", selection=True, stratify=True, missing=True):

        tic = time.perf_counter()
        print(f"{datetime.datetime.now()}: Simulation started.", flush=True)
        output_dict = self._traverse_graph(num_samples, output_path, missing, True)

        selectionNode = self._get_selection()
        if selection:
            if selectionNode is not None:
                print(f"{datetime.datetime.now()}: Simulating selection bias.", flush=True)
                output_dict = self.nodes[selectionNode]._filter_output(output_dict=output_dict)
                while len(list(output_dict.values())[0]) < num_samples:
                    temp_output = self.nodes[selectionNode]._filter_output(
                        output_dict=self._traverse_graph(1, output_path, missing, False))
                    output_dict = {k: output_dict[k] + temp_output[k] for k in output_dict.keys()}

        output_dict = {k: v for k, v in output_dict.items() if self.get_node_by_name(k).observed}
        output_dict = self._prettify_output(output_dict)

        stratifyNode = self._get_stratify()
        if stratify:
            if stratifyNode is not None:
                print(f"{datetime.datetime.now()}: Stratifying the data.", flush=True)
                output_dict = self.nodes[stratifyNode]._filter_output(output_dict=output_dict)

        if csv_name:
            if csv_name.endswith(".csv"):
                csv_name = csv_name[:-4]
            if stratifyNode is not None:
                for key in output_dict.keys():
                    pd.DataFrame(output_dict[key]).to_csv(output_path + csv_name + '_' + key + '.csv', index=False)
            else:
                pd.DataFrame(output_dict).to_csv(output_path + csv_name + '.csv', index=False)

        toc = time.perf_counter()
        print(f"{datetime.datetime.now()}: Simulation finished in {toc - tic:0.4f} seconds.", flush=True)
        return output_dict

    def _traverse_graph(self, num_samples, output_path, missing, show_log):
        output_dict = {}
        for node_name in self.top_order:

            node = self.get_node_by_name(node_name)
            if show_log and (isinstance(node, Missing) or isinstance(node, Node)):
                print(f"{datetime.datetime.now()}: Simulating node \"{node.name}\".", flush=True)
            node._node_simulate(num_samples, output_path)
            if isinstance(node, Selection):
                assert all(isinstance(x, bool) for x in node.output), "The selection node's function should return " \
                                                                      "a boolean"
            elif isinstance(node, Missing) and missing:
                assert all(isinstance(x, bool) for x in node.index_node.output), "The index node's function should " \
                                                                                 "return a boolean"
                output_dict[node.name] = node.output
            elif isinstance(node, Stratify):
                assert all(isinstance(x, str) for x in node.output), "The stratification node's function should " \
                                                                     "return a string"
            else:
                output_dict[node.name] = node.output
        return output_dict

    def _prettify_output(self, output_dict: dict):
        keys_to_remove = []
        for key in output_dict:
            node = self.get_node_by_name(key)
            # check if subscriptible
            if node.handle_multi_cols:
                node_output = output_dict[key]
                keys_to_remove.append(key)
                try:
                    unfolded_output = self._vec2dict(key, node_output)
                    output_dict = {**output_dict, **unfolded_output}
                except IndexError:
                    raise RuntimeError("All vectors returned by " + node.function.__name__ + " must be of the same "
                                                                                             "length")
                except TypeError:
                    raise RuntimeError("The output of " + node.function.__name__ + " either is not subscriptable "
                                                                                   "or has no __len__ method")
            elif node.handle_multi_return is not None:
                output_dict[key] = [node.handle_multi_return(elem) for elem in node.output]
        for key in keys_to_remove:
            output_dict.pop(key)

        return output_dict

    @staticmethod
    def _vec2dict(key: str, node_output):
        num_reps = len(node_output[0])
        node_dict = {key + "_" + str(rep): [] for rep in range(num_reps)}
        for sample in range(len(node_output)):
            for rep in range(num_reps):
                node_dict[key + "_" + str(rep)].append(node_output[sample][rep])
        return node_dict

    def ml_simulation(self, num_samples, train_test_ratio, stratify=False, include_external=False, csv_prefix=""):
        if csv_prefix:
            csv_prefix = csv_prefix + "_"
        num_tr_samples = int(num_samples * train_test_ratio)
        num_te_samples = num_samples - num_tr_samples
        train_dict = self.simulate(num_samples=num_tr_samples, stratify=stratify, csv_name=csv_prefix + "train")
        test_dict = self.simulate(num_samples=num_te_samples, stratify=stratify, csv_name=csv_prefix + "test")
        output = [train_dict, test_dict]
        if include_external:
            exter_dict = self.simulate(num_samples=num_te_samples, stratify=stratify, selection=False,
                                       csv_name=csv_prefix + "external")
            output.append(exter_dict)
        return output

    def _update_nodes(self, removed_nodes):
        # update the _constructors of the nodes to include the new parents
        for child_name in self.top_order:
            child = self.get_node_by_name(child_name)
            for parent in child.parents:
                usage = self.get_parent_usage(child, parent)
                if parent.name in removed_nodes:  # avoid modifying nodes in plates with parents not in a plate
                    if child.plates == parent.plates:  # todo change when you allow for nested plates
                        self._match_parents(child, parent, usage)
                    else:  # if child is in another plate, or not in a plate itfp
                        self._assign_parent_to_child(child, parent, usage)
                else:  #
                    if child.plates:
                        self._assign_parent_to_child(child, self.get_node_by_name(parent.name), usage, agg=0)
            child._args, child._kwargs = child._parse_func_arguments()
            child._update_parents()

    def get_parent_usage(self, child, parent):
        ind = child._constructor["args"].index(parent) if parent in child._constructor["args"] else None
        if ind is not None:
            tuple_usage = (ind, "arg")
        # assume each parent is used in one argument
        else:
            key = list(child._constructor["kwargs"].keys())[
                list(child._constructor["kwargs"].values()).index(parent)]
            tuple_usage = (key, "kwarg")
        assert tuple_usage is not None, f'{parent.name} is not a parent of {child.name}'
        return tuple_usage

    def _match_parents(self, child, parent, usage):
        replica_index = child.name.split("_")[-2]
        if usage[1] == "arg":
            child._constructor["args"][usage[0]] = self.get_node_by_name(f'{parent.name}_{replica_index}_')
        else:  # if in kwargs
            child._constructor["kwargs"][usage[0]] = self.get_node_by_name(f'{parent.name}_{replica_index}_')

    def _assign_parent_to_child(self, child, parent, usage, agg=1):
        parent_name = parent.name + "_agg" if agg else parent.name
        if usage[1] == "arg":
            child._constructor["args"][usage[0]] = self.get_node_by_name(parent_name)
        else:  # if in kwargs
            child._constructor["kwargs"][usage[0]] = self.get_node_by_name(parent_name)
